/********************************************************************\

  Name:         dsc.c
  Created by:   Stefan Ritt

  Contents:     Analyzer module for DSC decoding

  $Log$
\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include <math.h>

/* midas includes */
#include "midas.h"
#include "experim.h"
#include "analyzer.h"
#include "frontend.h"
#include "util.h"

/* cernlib includes */
#ifdef OS_WINNT
#define VISUAL_CPLUSPLUS
#endif

#ifdef OS_LINUX
#define f2cFortran
#endif

#include <cfortran.h>
#include <hbook.h>

/*-- Parameters ----------------------------------------------------*/

DSC_PARAM               dsc_param;
extern TRIGGER_SETTINGS trigger_settings;
extern GLOBAL_PARAM     global_param;

extern char             *analyzer_name;

/*-- Static parameters ---------------------------------------------*/

#define DSC_TEST_MODE        0
//#define DO_FIT
//#define STORE_FIT
#define DO_NORM

#define CALIB_PULSE_WIDTH  100      /* 100ns wide calibration pusle */
#define DSC_BASE          2000      /* 2000mV baseline              */
#define X_PEAK               9      /* peak in normalized waveform  */

#define N_PEAKS              6      /* save up to six peaks         */
#define N_PARAM      N_PEAKS*2      /* each peak has two params     */

#define DTDC_N_BINS        400
#define DTDC_X_LOW        -100
#define DTDC_X_HIGH        100

#define DADC_N_BINS        500
#define DADC_X_LOW           0
#define DADC_X_HIGH       2000

#define DSC_N_MODULES        1      /* number of 48 channel modules */

int dsc_n_channel[DSC_N_MODULES] = {
  16  /* Beam */
};

/*-- Module declaration --------------------------------------------*/

INT dsc_ana(EVENT_HEADER*, void*);
INT dsc_anainit(void);
INT dsc_bor(INT);
INT dsc_eor(INT);

DSC_PARAM_STR(dsc_param_str);

ANA_MODULE dsc_module = {
  "DSC",                         /* module name           */  
  "Stefan Ritt",                 /* author                */
  dsc_ana,                       /* event routine         */
  dsc_bor,                       /* BOR routine           */
  dsc_eor,                       /* EOR routine           */
  dsc_anainit,                   /* init routine          */
  NULL,                          /* exit routine          */
  &dsc_param,                    /* parameter structure   */
  sizeof(dsc_param),             /* structure size        */
  dsc_param_str,                 /* initial parameters    */
};

/*-- init routine --------------------------------------------------*/

WORD map[256];

float res[N_DSC];
int   n_res[N_DSC];

float ynorm[N_DSC][127];
int   ynormn[N_DSC][127];
float accum[N_DSC][127];
float taccum[N_DSC][127];
float accum_min[N_DSC];

INT dsc_anainit(void)
{
#ifdef DO_FIT
FILE *f;
int  j;
#endif

int  i;
char dsc_title[16][32] = {
  "DSC C0",
  "DSC C1",
  "DSC C2",
  "DSC C3",
  "DSC C4",
  "DSC C5",
  "DSC C6",
  "DSC C7",
  "DSC C8",
  "DSC C9",
  "DSC C10",
  "DSC C11", 
  "DSC C12",
  "DSC C13",
  "DSC C14",
  "DSC C15",
  };

  /* book raw DSC histos */
  for (i=0 ; i<N_DSC ; i++)
    HBOOK1(DSC_ID_BASE+i, dsc_title[i], 127, (float)0, (float)127, (float)0);

  /* initialize non-linear map */
  for (i=0 ; i<100 ; i++)
    map[i] = i;                   /* 0..100     -> 0..100     */
  for (i=100 ; i<200 ; i++)
    map[i] = ((i-100)*4+102);     /* 100..200   -> 100..500   */
  for (i=200 ; i<250 ; i++)
    map[i] = ((i-200)*10+505);    /* 200..250   -> 500..1000  */
  for (i=250 ; i<256 ; i++)            
    map[i] = ((i-250)*200+1100);  /* 250..255   -> 1000..2000 */

  /* load waveforms */
#ifdef DO_FIT
  f = fopen("waveform.dat", "r");
  if (f == NULL)
    cm_msg(MERROR, "dsc_bor", "cannot open file \"waveform.dat\"");
  else
    {
    for (i=0 ; i<N_DSC ; i++)
      for (j=0 ; j<127; j++)
        fscanf(f, "%f", &ynorm[i][j]);
    fclose(f);
    }
#endif

  return SUCCESS;
}

/*-- bor routine ---------------------------------------------------*/

INT dsc_bor(int run_number)
{
  memset(res, 0, sizeof(res));
  memset(n_res, 0, sizeof(n_res));

#ifdef DO_NORM
  memset(ynorm, 0, sizeof(ynorm));
  memset(ynormn, 0, sizeof(ynormn));
#endif

  return SUCCESS;
}

/*-- eor routine ---------------------------------------------------*/

INT dsc_eor(int run_number)
{
int    i, j;
char   str[80];
HNDLE  hDB, hkey;
FILE   *f;
float  norm, average;

  /*---- calculate average time resolution -------------------------*/

  average = norm = 0;
  for (i=0 ; i<N_DSC ; i++)
    {
    if (n_res[i] > 100)
      {
      dsc_param.ns_per_bin[i] = res[i]/n_res[i];
      average += dsc_param.ns_per_bin[i];
      norm++;
      }
    }
  if (norm > 0)
    average /= norm;

  /* set ODB values */
  cm_get_experiment_database(&hDB, NULL);
  sprintf(str,"/%s/Parameters/DSC", analyzer_name);
  db_find_key(hDB, 0, str, &hkey);
  db_set_value(hDB, hkey, "ns per bin", dsc_param.ns_per_bin, 
               sizeof(float)*N_DSC, N_DSC, TID_FLOAT);

  /*---- caluclate normalized waveform and save --------------------*/

#ifdef DO_NORM
  f = fopen("waveform.dat", "w");
  if (f == NULL)
    cm_msg(MERROR, "dsc_eor", "cannot open file %s", str);
  else
    {
    for (i=0 ; i<N_DSC ; i++)
      {
      for (j=0 ; j<127; j++)
        {
        if (ynormn[i][j] > 0)
          ynorm[i][j] /= ynormn[i][j];
        }
      if (ynorm[i][X_PEAK] == 0)
        norm = 1;
      else
        norm = ynorm[i][X_PEAK];

      // printf("Channel %d has %d hits\n", i, ynormn[i][X_PEAK]);

      for (j=0 ; j<120; j++)
        {
        /* reverse and normalize function from 0...1 */
        fprintf(f, "%1.3f ", ynorm[i][j] / norm);
        }

      /* extrapolate last 6 bins */
      for (j=110,average=0 ; j<120 ; j++)
        average += ynorm[i][j];
      average /= 10;

      for (j=120 ; j<127 ; j++)
        fprintf(f, "%1.3f ", average / norm);
      fprintf(f, "\n");
      }
    fclose(f);
    }
#endif

  return SUCCESS;
}

/*-- fit routine ---------------------------------------------------*/

int f_channel, f_n_peak;

double f_dsc(double x, double param[])
{
int    i, j;
double y, sum, x1;

  sum = 0;

  for (i=0 ; i<f_n_peak ; i++) 
    {
    x1 = x - param[i*2+1] + X_PEAK;
    j = (int) x1;

    if (j >= 0 && j < 126)
      {
      y = ynorm[f_channel][j] + (x1-j)*(ynorm[f_channel][j+1] -
                                        ynorm[f_channel][j]);
      y *= param[i*2];

      sum += y;
      }
    }

  return sum;
}

/*-- event routine -------------------------------------------------*/

INT dsc_ana(EVENT_HEADER *pheader, void *pevent)
{
int           i, j, k, i1, i2, n_rdsc, n_cdsc, channel, n_peak;
int           range, imax, imaxgl;
WORD          *rdsc;
BYTE          *cdsc;
float         *y, yfloat[N_DSC*132], raw[N_DSC*132], x1, x2, ymin;
float         ymingl, ylimit;
static DWORD  last_time = 0;
static double x[127];
BOOL          channel_map[N_DSC];
float         ysmooth[127], ytmp[127], ysd[127], ybase, ysum;

#ifdef DO_FIT
int           index;
BOOL          protect[127];
double        param[N_PARAM], error[N_PARAM], chisqr, 
              yd[127], yfit[127], weight[127];
float         peak_time[N_PEAKS], peak_adc[N_PEAKS];
float         ydiff, *pf;
float         dsc_time[N_DSC], dsc_adc[N_DSC];
WORD          dsc_nhit[N_DSC], *pw;
#endif;

  /* look for RDSC bank, return if not present */
  n_rdsc = bk_locate(pevent, "RDSC", &rdsc);
  n_cdsc = bk_locate(pevent, "CDSC", &cdsc);
  if (rdsc == NULL && cdsc == NULL)
    return SUCCESS;

  /*---- unpack data -----------------------------------------------*/

  if (rdsc != NULL)
    {
    for (i=0 ; i<N_DSC*132 ; i++)
      yfloat[i] = rdsc[i];
    }

  if (cdsc != NULL)
    {
    for (i=0 ; i<N_DSC*132 ; i++)
      yfloat[i] = DSC_BASE;

    i = 0;
    if (n_cdsc > 0)
      {
      do
        {
        channel = cdsc[i] | ((cdsc[i+1] & 3) << 8);
        n_peak = cdsc[i+1] >> 2;
        i += 2;
        channel_map[channel] = TRUE;

        for (j=0 ; j<n_peak ; j++)
          {
          i1 = cdsc[i++];
          i2 = cdsc[i++];
          ymin = DSC_BASE;

          /* fill yfloat array */
          for (k = i1 ; k<i2 ; k++)
            {
            yfloat[channel*132+k] = (float) (DSC_BASE - map[cdsc[i++]]);
            if (yfloat[channel*132+k] < ymin)
              ymin = yfloat[channel*132+k];
            }

          }
        } while (i<n_cdsc);
      }  
    }

  /*---- write raw waveform once every second to histo -------------*/

  if (ss_time() - last_time > 1)
    {
    last_time = ss_time();

    for (i=0 ; i<N_DSC ; i++)
      HPAK(DSC_ID_BASE+i, &yfloat[i*132+1]);
    }

  /*---- calibrate time scale with random events -------------------*/

  if (cdsc != NULL && DSC_TEST_MODE)
    {
    for (channel=0 ; channel<N_DSC ; channel++)
      {
      y = yfloat+channel*132;

      for (i=i1=0,ymin=DSC_BASE ; i<60 ; i++)
        if (y[i] < ymin) 
          {
          ymin = y[i];
          i1 = i;
          }
      
      for (i=i2=60,ymin=DSC_BASE ; i<120 ; i++)
        if (y[i] < ymin) 
          {
          ymin = y[i];
          i2 = i;
          }

      if (y[i1] < 1970 && y[i2] < 1970) 
        {  
        /* fit parabola through three points around maximum */
        x1 = (y[i1-1]-y[i1+1]) /
             (2*y[i1-1]+2*y[i1+1]-4*y[i1]) + i1;
        x2 = (y[i2-1]-y[i2+1]) /
             (2*y[i2-1]+2*y[i2+1]-4*y[i2]) + i2;
           
        /* cut on overlaid real pulses (especially for beam counters) */
        if (x2-x1 > 45 && fabs(y[i2]-y[i1]) < 30)
          {
          res[channel] += CALIB_PULSE_WIDTH/(x2-x1);
          n_res[channel]++;
          }
        }
      }
    }

  /*---- normalize waveform for raw events -------------------------*/

  if (rdsc != NULL && DSC_TEST_MODE)
    {
    for (channel=0 ; channel<N_DSC ; channel++)
      {
      y = raw+channel*132+1;

      /* do smoothing, skip three bins from first/last */
      for (i=0 ; i<127 ; i++)
        ysmooth[i] = y[i];
      for (j=0 ; j<2 ; j++) 
        {
        ytmp[3] = (2*ysmooth[3] + ysmooth[4])/3;
        ytmp[124] = (2*ysmooth[124] + ysmooth[123])/3;
        for (i=4 ; i<124 ; i++)
          ytmp[i] = (float) (0.25*(ysmooth[i-1] + 2*ysmooth[i] + ysmooth[i+1]));
        for (i=3 ; i<125 ; i++)
          ysmooth[i] = ytmp[i];
        }
      ysmooth[0] = ysmooth[1] = ysmooth[2] = ysmooth[3];
      ysmooth[127] = ysmooth[126] = ysmooth[125] = ysmooth[124];

      /* evaluate baseline */
      range = 5;
      for (i=0,ybase=0,imax=0 ; i<127 ; i++)
        if (ysmooth[i] > ybase) 
          {
          ybase = ysmooth[i];
          imax = i;
          }
      if (imax < range)
        imax = range;
      if (imax > 126-range)
        imax = 126-range;
    
      for (i=imax-range,ysum=0,j=0 ; i<=imax+range ; i++) 
        {
        if (ysmooth[i] > ybase-15) 
          {
          ysum += ysmooth[i];
          j++;
          }
        }
      ybase = ysum/j;
    
      /* calculate second derivative */
      for (i=0 ; i<127 ; i++) 
        {
        i1 = i<3 ? i+3 : i-3;
        i2 = i>=127-5 ? i-5 : i+5;
    
        ysd[i] = 2*ysmooth[i] - ysmooth[i1] - ysmooth[i2];
        }

      /* search peaks */
      ymingl = ylimit = 0;
      n_peak = imaxgl = 0;
      do 
        {
        /* find maximum in second derivative, omit first and last three bins */  
        for (i=3,ymin=0 ; i<124 ; i++)
          if (ysd[i] < ymin) 
            {
            ymin = ysd[i];
            imax = i;
            }
    
        /* remember global minimum */
        if (ymingl == 0) 
          {
          ymingl = ymin;
      
          /* use limit=global min/10 or -20 */  
          ylimit = min(ymingl/10, -20);  
          }
    
        /* check if peak is above limit */
        if (ymin < ylimit && ysmooth[imax] < ybase-10) 
          {
          n_peak++;
          imaxgl = imax;

          /* delete region around peak */
          for (i=max(0,imax-3) ; i<min(127,imax+5) ; i++)
            ysd[i] = 0;
          } 
        else
          break;

        } while (n_peak < N_PEAKS);
      
      /* only normalize if only one peak */
      if (n_peak == 1 && ybase - ysmooth[imaxgl] > 30) 
        {
        /* find peak minimum */
        i1 = max(0, imaxgl-3);
        i2 = min(126, imaxgl+3);
        
        for (i=i1,ymin=ysmooth[i1] ; i<=i2 ; i++)
          if (ysmooth[i] < ymin)
            {
            ymin = ysmooth[i];
            imaxgl = i;
            }
          
        for (i=0 ; i<127 ; i++) 
          {
          i1 = i+imaxgl-X_PEAK;
          if (i1 >= 0 && i1 < 127) 
            {
            ynorm[channel][i] += ybase-ysmooth[i1];
            ynormn[channel][i]++; 
            }            
          }
        }
      }
    }

  /*---- fit waveforms ---------------------------------------------*/

#ifdef DO_FIT

  /* initialize arrays */
  if (x[0] == 0)
    for (i=0 ; i<127 ; i++)
      x[i] = i;

  memset(dsc_nhit, 0, sizeof(dsc_nhit));
  memset(dsc_adc, 0, sizeof(dsc_adc));
  for (i=0 ; i<N_DSC ; i++)
    dsc_time[i] = -1000.f;

  if (cdsc != NULL && DSC_TEST_MODE) == 0)
    {
#ifdef STORE_FIT    
    bk_create(pevent, "DFIT", TID_FLOAT, &pf);
    memset(pf, 0, sizeof(float)*N_DSC*127);
#endif

    for (channel=0 ; channel<N_DSC ; channel++)
      {
      if (!channel_map[channel])
        continue;

      y = yfloat+channel*132;
      f_channel = channel;
      memset(protect, 0, sizeof(protect));

      /* search peaks */
      ymingl = ylimit = 0;
      n_peak = imaxgl = 0;
      do 
        {
        f_n_peak = n_peak;

        /* find signal minimum */  
        for (i=0,ymin=0 ; i<127 ; i++)
          {
          ydiff = (float) (y[i]-DSC_BASE+f_dsc(x[i], param));

          if (ydiff < ymin && !protect[i]) 
            {
            ymin = ydiff;
            imax = i;
            }
          }
    
        /* check if peak is significant */
        if (ymin > -10)
          break;

        /* set initial fit parameters on peak */
        param[n_peak*2+0] = DSC_BASE-y[imax]-f_dsc(x[imax], param);
        param[n_peak*2+1] = imax;

        /* protect area around peak */
        for (i=max(0, imax-5) ; i<min(127,imax+5) ; i++)
          protect[i] = TRUE;

        n_peak++;

        } while (n_peak < N_PEAKS);

      if (n_peak > 0)
        {
        for (i=0 ; i<127 ; i++)
          {
          yd[i] = DSC_BASE - y[i];
          weight[i] = (y[i] == DSC_BASE) ? 0 : 1;
          }

refit:
        f_n_peak = n_peak;
        memset(peak_time, 0, sizeof(peak_time));
        memset(peak_adc, 0, sizeof(peak_adc));

        chisqr = fit(x, yd, weight, f_dsc, 127, n_peak*2, param, error, yfit);
        
        /*
        for (i=0 ; i<127 ; i++)
          yfit[i] = f_dsc(x[i], param);
        chisqr = 1;
        */
        
        if (chisqr != -1)
          {
          for (i=0 ; i<n_peak ; i++)
            {
            /* if peak has been fitted away, remove it and fit again */
            if (param[i*2] < 0 || param[i*2+1] < -30 || param[i*2+1]>127+30)
              {
              /*
              printf("remove ser %d chn %d n %d\n", pheader->serial_number, channel, n_peak);
              */

              for (j=i+1 ; j<n_peak ; j++)
                {
                param[i*2] = param[j*2];
                param[i*2+1] = param[j*2+1];
                }
              n_peak--;
              if (n_peak == 0)
                break;

              goto refit;
              }

            /* remove small peaks for degrader since they are fake */

            /* not true for new data ...
            if (param[i*2] < 15 && channel == 10)
              {
              for (j=i+1 ; j<n_peak ; j++)
                {
                param[i*2] = param[j*2];
                param[i*2+1] = param[j*2+1];
                }
              n_peak--;
              if (n_peak == 0)
                break;

              goto refit;
              }
            */

            /* calibrate time and gain */
            peak_time[i] = (float) param[i*2+1]*dsc_param.ns_per_bin[channel] - 
                           dsc_param.tdc_offset[channel];
            peak_adc[i]  = (float) param[i*2] * dsc_param.adc_gain[channel];

/*
          if (channel == 10 && n_peak > 3)
            printf("%5d-%2d: %5.1lf %5.1lf %5.2lf\n", pheader->serial_number, 
                   channel, peak_adc[i], peak_time[i], chisqr);
*/
            }

          dsc_nhit[channel] = n_peak;

          if (n_peak > 0)
            {
            /* sort hits time-wise (bubble-sort) */
            for (i=0 ; i<n_peak-1 && peak_adc[i+1] > 0; i++)
              for (j=i+1 ; j<n_peak && peak_adc[j] > 0; j++)
                if (peak_time[j] < peak_time[i])
                  {
                  float tmp;

                  tmp = peak_time[i];
                  peak_time[j] = peak_time[i];
                  peak_time[i] = tmp;

                  tmp = peak_adc[i];
                  peak_adc[j] = peak_adc[i];
                  peak_adc[i] = tmp;
                  }

            /* store peaks in pre/post variables */
            for (i=index=0 ; i<n_peak ; i++)
              if (fabs(peak_time[i]) < fabs(dsc_time[channel]))
                {
                dsc_time[channel] = peak_time[i];
                dsc_adc[channel] = peak_adc[i];
                index = i;
                }

#ifdef STORE_FIT
            for (i=0 ; i<127 ; i++)
              pf[channel*127+i] = (float) ((int)(10*yfit[i]+0.5)/10);

            /* store peak maximum in float number */
            for (i=0 ; i<n_peak ; i++)
              if (param[i*2+1] > 0)
                pf[channel*127+(int) (param[i*2+1]+0.5)] += 0.05f;
#endif
            }
          }
        }
      }

#ifdef STORE_FIT
    bk_close(pevent, pf+N_DSC*127);
#endif

    }

  bk_create(pevent, "DHIT", TID_WORD, &pw);
  for (i=0 ; i<N_DSC ; i++)
    *pw++ = dsc_nhit[i];
  bk_close(pevent, pw);

  bk_create(pevent, "DTIM", TID_FLOAT, &pf);
  for (i=0 ; i<N_DSC ; i++)
    *pf++ = dsc_time[i];
  bk_close(pevent, pf);

  bk_create(pevent, "DADC", TID_FLOAT, &pf);
  for (i=0 ; i<N_DSC ; i++)
    *pf++ = dsc_adc[i];
  bk_close(pevent, pf);

#endif /* DO_FIT */

  return SUCCESS;
}
